#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "ttypes.h"

#include "ttheader.h"
#include "ttdir.h"
#include "ttname.h"
#include "ttos2.h"

static const char *font_styles[4]={"medium","bold","medium","bold"};
static const char *font_style_bit[4]={"r","r","i","i"};
static const char *font_spacing_bit[2]={"p","m"};

static int font_spacing;

static int j,k;
static FILE *f;
static char eid[5];

static struct _ttf_directory 		ttf_dir;
static struct _ttf_directory_entry 	ttf_entry;
static struct _ttf_header		ttf_header;
static struct _ttf_name_header		ttf_name_header;
static struct _ttf_name_record		ttf_name_record;
static struct _ttf_os2_header		ttf_os2_header;

/*static long fpos;*/
static long headpos;
static long namepos;
static long os2pos;

int n_fonts;
static int name_platform;
char *name;


struct _font_dir_string {
	char *string;
	struct _font_dir_string *next;
};

char *astring;
struct _font_dir_string *font_dir_string_first=NULL;
struct _font_dir_string *font_dir_string_current=NULL;




static void swap_2(unsigned short *x)
{
unsigned short y;
	
	y=(*x>>8)+(*x<<8);
	*x=y;
}

static void swap_4(unsigned long *x)
{
unsigned long y;
	
	y=(((*x)&0x000000FF)<<24) + (((*x)&0x0000FF00)<<8) + (((*x)&0x00FF0000)>>8) + (((*x)&0xFF000000)>>24);
	*x=y;
}



static int read_directory(void)
{
	if (fread(&ttf_dir,sizeof(ttf_dir),1,f)<1) return 0;

	/* of course its got the wrong byte order.... sigh */
	swap_2(&ttf_dir.version.major);
	swap_2(&ttf_dir.version.minor);
	swap_2(&ttf_dir.entries);
	swap_2(&ttf_dir.search_range);
	swap_2(&ttf_dir.entry_selector);
	swap_2(&ttf_dir.range_shift);

	return 1;
}


static int read_entry(void)
{
	if (fread(&ttf_entry,sizeof(ttf_entry),1,f)<1) return 0;
	
	swap_4(&ttf_entry.checksum);
	swap_4(&ttf_entry.offset);
	swap_4(&ttf_entry.length);
	
	return 1;
}



static int read_header(void)
{
	if (!headpos) return 0;

	fseek(f,headpos,SEEK_SET);
	if (fread(&ttf_header,sizeof(ttf_header),1,f)<1) return 0;

	swap_2(&ttf_header.table_version.major);
	swap_2(&ttf_header.table_version.minor);
	swap_2(&ttf_header.font_revision.major);
	swap_2(&ttf_header.font_revision.minor);
	swap_4(&ttf_header.checksum);
	swap_4(&ttf_header.magic);
	swap_2(&ttf_header.flags);
	swap_2(&ttf_header.units_per_em);
	swap_2(&ttf_header.xmin);
	swap_2(&ttf_header.ymin);
	swap_2(&ttf_header.xmax);
	swap_2(&ttf_header.ymax);
	swap_2(&ttf_header.style);
	swap_2(&ttf_header.lowest_rec_ppem);
	swap_2(&ttf_header.font_direction_hint);
	swap_2(&ttf_header.index_format);
	swap_2(&ttf_header.glyphdata_format);

	return 1;
}







static int read_name_header(void)
{
	if (!namepos) return 0;

	fseek(f,namepos,SEEK_SET);
	if (fread(&ttf_name_header,sizeof(ttf_name_header),1,f)<1) return 0;
	swap_2(&ttf_name_header.format);
	swap_2(&ttf_name_header.records);
	swap_2(&ttf_name_header.strings_offset);
	
	return 1;
}



static int read_name_record(int recno)
{
	if (!namepos) return 0;
	
	fseek(f,namepos+sizeof(ttf_name_header)+(sizeof(ttf_name_record)*recno),SEEK_SET);
	if (fread(&ttf_name_record,sizeof(ttf_name_record),1,f)<1) return 0;
	swap_2(&ttf_name_record.platform_id);
	swap_2(&ttf_name_record.encoding_id);
	swap_2(&ttf_name_record.language_id);
	swap_2(&ttf_name_record.name_id);
	swap_2(&ttf_name_record.name_length);
	swap_2(&ttf_name_record.name_offset);
	
	return 1;
}




static char *read_name(unsigned long string_offset, int length, int bytes_per_char)
{
char *s;
int i;	
	
	if (!namepos) return NULL;
	fseek(f,namepos+ttf_name_header.strings_offset+string_offset,SEEK_SET);

	s=(char *)malloc(length+1);
	for (i=0; i<length; i++) {
		if (bytes_per_char>1) fseek(f,bytes_per_char-1,SEEK_CUR);
		if (fread(&s[i],1,1,f)<1) {
			free(s);
			return NULL;
		}
	}
	s[length]=0;

	return s;
}





static int read_os2_header(void)
{
int i;

	if (!os2pos) return 0;

	fseek(f,os2pos,SEEK_SET);
	if (fread(&ttf_os2_header,sizeof(ttf_os2_header),1,f)<1) return 0;
	swap_2(&ttf_os2_header.version);
	swap_2(&ttf_os2_header.avg_char_width);
	swap_2(&ttf_os2_header.weight_class);
	swap_2(&ttf_os2_header.width_class);
	swap_2(&ttf_os2_header.fs_type);
	swap_2(&ttf_os2_header.subscript_x_size);
	swap_2(&ttf_os2_header.subscript_y_size);
	swap_2(&ttf_os2_header.subscript_x_offset);
	swap_2(&ttf_os2_header.subscript_y_offset);
	swap_2(&ttf_os2_header.superscript_x_size);
	swap_2(&ttf_os2_header.superscript_y_size);
	swap_2(&ttf_os2_header.superscript_x_offset);
	swap_2(&ttf_os2_header.superscript_y_offset);
	swap_2(&ttf_os2_header.strikeout_size);
	swap_2(&ttf_os2_header.strikeout_pos);
	swap_2(&ttf_os2_header.family_class);
	for (i=0; i<4; i++)
		swap_4(&ttf_os2_header.char_range[i]);
	swap_2(&ttf_os2_header.selection);
	swap_2(&ttf_os2_header.first_char_index);
	swap_2(&ttf_os2_header.last_char_index);
	swap_2(&ttf_os2_header.typo_ascender);
	swap_2(&ttf_os2_header.typo_descender);
	swap_2(&ttf_os2_header.typo_gap);
	swap_2(&ttf_os2_header.win_ascend);
	swap_2(&ttf_os2_header.win_descend);
	return 1;
}





static void add_dir_string(char *string)
{
struct _font_dir_string * stmp;
	
	stmp=font_dir_string_current;
	font_dir_string_current=(struct _font_dir_string *)malloc(sizeof(struct _font_dir_string));
	if (!font_dir_string_first) {
		font_dir_string_first=font_dir_string_current;
	}
	if (stmp) stmp->next=font_dir_string_current;
	font_dir_string_current->next=NULL;
	font_dir_string_current->string=string;
}


static void free_all_dir_strings(void)
{
struct _font_dir_string *stmp;
	
	if (!font_dir_string_first) return;

	font_dir_string_current=font_dir_string_first;
	while (font_dir_string_current->next) {
		stmp=font_dir_string_current->next;
		free(font_dir_string_current);
		font_dir_string_current=stmp;
	}
	
}




int main(int argc, char *argv[])
{
int bytes_per_char;
int i;
	
	if (argc<2) {
		printf("mkttfontdir v1.0. Public domain software.\n");
		printf("\nUsage : mkttfontdir <font1.ttf> [font2.ttf] [...]\n");
		printf("But usually just : ./mkttfontdir *.ttf\n");
		printf("mkttfontdir will generate a fonts.dir file for use with X11 that contains entries for the\n");
		printf("ttf fonts you specify on the command-line.\n\n");
		exit(0);
	}

	n_fonts=0;
	for (i=1; i<argc; i++) {
		f=fopen(argv[i],"rb");
		if (!f) {
			fprintf(stderr,"Cannot open file %s for reading.\n",argv[i]);
			goto nextfile;
		} 

		if (!read_directory()) {
			fprintf(stderr,"Read error in file %s.\n",argv[i]);
			goto nextfile;
		}
		
	        headpos=0;
		namepos=0;
		
		for (j=0; j<ttf_dir.entries; j++) {
			if (!read_entry()) {
				fprintf(stderr,"Table entry read error in file %s.\n",argv[i]);
				goto nextfile;
			}
			strncpy(eid,ttf_entry.id,4);
			eid[4]=0;
			if (strcmp(eid,"head")==0) headpos=ttf_entry.offset;
			if (strcmp(eid,"name")==0) namepos=ttf_entry.offset;
			if (strcmp(eid,"OS/2")==0) os2pos=ttf_entry.offset;
			
		}
		
		if (!read_header()) {
			fprintf(stderr,"Read error in file %s. Cannot read font header.\n",argv[i]);
			goto nextfile;
		} 
		
		if (!read_name_header()) {
			fprintf(stderr,"Read error in file %s. Cannot read name header.\n",argv[i]);
			goto nextfile;
		} else {

			name=NULL;
			name_platform=-1;
			
			for (j=0; j<ttf_name_header.records; j++) {
				if (!read_name_record(j)) {
					fprintf(stderr,"Read error in file %s. Cannot read name record %d.\n",argv[i],j);
				} else {

					if ((name) && (ttf_name_record.platform_id>name_platform)) {
						free(name);
						name=NULL;
					}
					if ((ttf_name_record.name_id==1) && (!name)) {
						bytes_per_char=1;
						if (ttf_name_record.platform_id==3) bytes_per_char=2;
						name=read_name(ttf_name_record.name_offset,ttf_name_record.name_length/bytes_per_char,bytes_per_char);
						name_platform=ttf_name_record.platform_id;
						if (!name) {
							fprintf(stderr,"Read error in file %s. Cannot read font name.\n",argv[i]);
						}
					}
					
				}
			}
			
		}

		if (!read_os2_header()) {
			font_spacing=0;
		} else {
			if (ttf_os2_header.panose[3]==9)
				font_spacing=1;
			else
				font_spacing=0;
		}

		if (name) {
			if (strlen(name)==0) {
				free(name);
				name=(char *)malloc(strlen(argv[i])+1);
				strcpy(name,argv[i]);
			}
			k=strlen(name);
			for (j=0; j<k; j++) {
				name[j]=tolower(name[j]);
			}

			astring=(char *)malloc(strlen(argv[i])+strlen(name)+strlen(font_styles[ttf_header.style&3])+strlen(font_style_bit[ttf_header.style&3])+strlen(font_spacing_bit[font_spacing])+50);
			sprintf(astring,"%s -truetype-%s-%s-%s-normal-0-0-0-0-0-%s-0-iso8859-1",argv[i],name,font_styles[ttf_header.style&3],font_style_bit[ttf_header.style&3],font_spacing_bit[font_spacing]);
			add_dir_string(astring);
			n_fonts++;
			free(name);
		} else {
			printf("Font name not found in font %s.\n",argv[i]);
		}
		
		fclose(f);

nextfile:
	}

	if (n_fonts>0) {
		f=fopen("fonts.dir","wb");
		if (!f) {
			fprintf(stderr,"Cannot open fonts.dir for writing.\nMake sure you have write permissions, or run as root.\n");
			return 1;
		}
		
		fprintf(f,"%d\n",n_fonts);
		
		font_dir_string_current=font_dir_string_first;
		for (i=0; i<n_fonts; i++) {
			fprintf(f,"%s\n",font_dir_string_current->string);
			font_dir_string_current=font_dir_string_current->next;
		}

		fclose(f);
		
		free_all_dir_strings();
	}
	
	switch(n_fonts) {
		default:
			printf("%d fonts found and installed.\n",n_fonts);
			break;
		
		case 0:
			printf("No fonts found.\n");
			break;
		
		case 1:
			printf("1 font found and installed.\n");
			break;
	}
	
	return 0;
}


